package nsf;

import java.io.BufferedWriter;
import java.util.ArrayList;
import java.io.File;
import java.util.Collections;
import java.util.Enumeration;
import java.util.Hashtable;
import org.xml.sax.Attributes;
import org.xml.sax.Locator;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import org.xml.sax.ext.DefaultHandler2;

import javax.xml.parsers.SAXParserFactory;
import javax.xml.parsers.SAXParser;
import org.sanskritlibrary.IOAssistance;

public class ParseTEI extends DefaultHandler2 //implements ErrorHandler, ContentHandler, DTDHandler, EntityResolver
{
    static public final String rowTemplate = "<tr><td valign='top'>%s:</td><td>%s</td></tr>\n";
    static public final int ACTION_INSERT = 1;
    static public final int ACTION_FORMAT = 2;
    static public final int ACTION_EXTRACT = 3;
    static private PathInfo pi;
    static private ParseDocument pd;
    static private Hashtable<String, Integer> allpaths = new Hashtable<String, Integer>();

    int workNumber = 1;
    String currentDirectory;
    Locator locator;
    ArrayList<String> path = new ArrayList<String>();
    int actionType;
    int debug = 0;
    int ignoreUntil = -1;
    BufferedWriter bw;
    String dir = "Catalogs/Manuscripts/";
    boolean endCommentWithLineFeed = true;      // for comments before root element
    String extractionPath;
    StringBuilder works = new StringBuilder();
    StringBuilder manuscriptIdentifiers = new StringBuilder();
    StringBuilder physicalDescription = new StringBuilder();

    StringBuilder hands = new StringBuilder();
    ArrayList<String> handNotes = new ArrayList<String>();

    StringBuilder decoration = new StringBuilder();
    StringBuilder history = new StringBuilder();

    StringBuilder subjectHeadings = new StringBuilder();
    ArrayList<String>subjectItems = new ArrayList<String>();

    StringBuilder facsimile = new StringBuilder();
    ArrayList<String> facsimileItems = new ArrayList<String>();

    StringBuilder transcription = new StringBuilder();

    ArrayList<ChangeInfo>changeItems = new ArrayList<ChangeInfo>();
    StringBuilder changes = new StringBuilder();

    String theFileName;

    public ParseTEI(String fileName, int actionType) throws Exception
    {
        this(fileName, actionType, null, null);
    }

    public ParseTEI(String fileName, int actionType, String extractionPath, BufferedWriter bw) throws Exception
    {
        File fn = new File(fileName);
        String name = fn.getName();
        theFileName = name;             // for debugging
        this.actionType = actionType;
        this.extractionPath = extractionPath;
        this.bw = bw;

        if (actionType == ACTION_FORMAT/* && pi == null*/)
        {
            pd = new ParseDocument();
            pd.setDefaultTagHandler(new Rewrite_default());

            pi = new PathInfo();

            reserveEntries();

            // Identify containers (will cause xml to be copied as is into buffer for path)
            pi.container("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.additions");
            pi.container("TEI.teiHeader.fileDesc.sourceDesc.msDesc.history.origin");
            pi.container("TEI.teiHeader.fileDesc.sourceDesc.msDesc.history.provenance");
            pi.container("TEI.teiHeader.fileDesc.sourceDesc.msDesc.history.acquisition");
            pi.container("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.bindingDesc");
            pi.container("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.sealDesc");
            pi.container("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.handDesc.handNote");
            pi.container("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.decoDesc.decoNote[type='color']");
            pi.container("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.decoDesc.decoNote[type='border']");
            pi.container("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.decoDesc.decoNote[type='illustration']");
            pi.container("TEI.text");


            // Set keys
            pi.key("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msIdentifier.altIdentifier", "type");
            pi.key("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msContents.msItemStruct.title", "type");
            pi.key("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msContents.msItemStruct.rubric", "type");
            pi.key("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msIdentifier.altIdentifier.collection", "type");
            pi.key("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.decoDesc.decoNote", "type");
            pi.key("TEI.teiHeader.profileDesc.textClass.keywords", "scheme");

            // Set handlers (will be called when end tag of path is reached)
            pi.handler("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc", new Handler_physDesc());
            pi.handler("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msIdentifier", new Handler_msIdentifierStruct());
            pi.handler("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msContents.msItemStruct", new Handler_msItemStruct());
            pi.handler("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.handDesc.handNote", new Handler_handNote());
            pi.handler("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.additions", new Handler_handDesc());
            pi.handler("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.decoDesc", new Handler_decoDesc());
            pi.handler("TEI.teiHeader.fileDesc.sourceDesc.msDesc.history", new Handler_history());
            pi.handler("TEI.teiHeader.profileDesc.textClass.keywords[scheme='#LCSH'].list.item", new Handler_profileDescItem());
            pi.handler("TEI.teiHeader.profileDesc", new Handler_profileDesc());
            pi.handler("TEI.facsimile.graphic", new Handler_facsimileGraphic());
            pi.handler("TEI.facsimile", new Handler_facsimile());
            pi.handler("TEI.teiHeader.fileDesc.sourceDesc.msDesc.additional.adminInfo.recordHist.change", new Handler_change());
            pi.handler("TEI.teiHeader.fileDesc.sourceDesc.msDesc.additional.adminInfo.recordHist", new Handler_changes());
            pi.handler("TEI.text", new Handler_transcription());
            pi.setDefaultHandler(new Handler_default());
        }

        // Use the default (non-validating) parser
        SAXParserFactory factory = SAXParserFactory.newInstance();
        try
        {
            // Parse the input
            SAXParser saxParser = factory.newSAXParser();
            if (actionType == ACTION_INSERT)
            {
                // See http://www.saxproject.org/apidoc/org/xml/sax/package-summary.html#package_description,
                saxParser.setProperty("http://xml.org/sax/properties/lexical-handler", this);
                this.bw = IOAssistance.openWriter(dir + name);
                this.bw.write("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
            }
            else if (actionType == ACTION_FORMAT)
            {
                name = name.replace(".xml", ".html");
                this.bw = IOAssistance.openWriter(dir + name);
                this.bw.write(String.format("<html><head><title>%s</title></head><body>\n", name));
            }
            else if (actionType == ACTION_EXTRACT)
            {
                // See http://www.saxproject.org/apidoc/org/xml/sax/package-summary.html#package_description,
                endCommentWithLineFeed = true;
                saxParser.setProperty("http://xml.org/sax/properties/lexical-handler", this);
            }

            saxParser.parse(fn, this);
        }
        catch (Throwable t)
        {
            t.printStackTrace ();
        }
    }


    public static void writeAllPaths(BufferedWriter bw) throws Exception
    {
        Enumeration<String> en = allpaths.keys();
        ArrayList<TagCount> a = new ArrayList<TagCount>();
        while (en.hasMoreElements())
        {
            String path = en.nextElement();
            Integer cnt = allpaths.get(path);
            a.add(new TagCount(path, cnt));
        }

        Collections.sort(a, new TagCount());

        for (TagCount tc : a)
        {
            bw.write(String.format("%s (%d)\n", tc.path, tc.count));
        }
    }

    private String getPath()
    {
        StringBuilder sb = new StringBuilder();
        String delim = "";
        for (String name : path)
        {
            sb.append(String.format("%s%s", delim, name));
            delim = ".";
        }

        return sb.toString();
    }

    
    private void pushPath(String qname)
    {
        path.add(qname);
    }

    private String popPath()
    {
        int top = path.size()-1;
        String s = path.get(top);
        path.remove(top);
        return s;
    }

    // For ErrorHandler
    @Override
    public void warning(SAXParseException exception) throws SAXException
    {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public void error(SAXParseException exception) throws SAXException
    {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public void fatalError(SAXParseException exception) throws SAXException
    {
        throw exception;
    }

    // For ContentHandler
    @Override
    public void setDocumentLocator(Locator locator)
    {
        this.locator = locator;
    }

    @Override

    public void startDocument() throws SAXException
    {
    }

    @Override
    public void endDocument() throws SAXException
    {
        if (actionType == ACTION_FORMAT || actionType == ACTION_INSERT)
        {
            try
            {
                if (actionType == ACTION_FORMAT)
                {
                    bw.write(manuscriptIdentifiers.toString());
                    bw.write(works.toString());
                    bw.write(physicalDescription.toString());
                    bw.write(hands.toString());
                    bw.write(decoration.toString());
                    bw.write(history.toString());
                    bw.write(subjectHeadings.toString());
                    bw.write(facsimile.toString());
                    bw.write(transcription.toString());
                    bw.write(changes.toString());
                    bw.write("</body></html>");
                }
                bw.close();
            }
            catch (Exception ex)
            {
                throw new SAXException(ex.getMessage());
            }            
        }
    }

    @Override
    public void startPrefixMapping(String prefix, String uri) throws SAXException
    {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public void endPrefixMapping(String prefix) throws SAXException
    {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public void startElement(String uri, String localName, String qName, Attributes atts) throws SAXException
    {
        debug++;
        if (debug == 185)
            debug = 185;
        try
        {
            if (actionType == ACTION_INSERT)
                startElementInsert(uri, localName, qName, atts);
            else if (actionType == ACTION_FORMAT)
                startElementFormat(uri, localName, qName, atts);
            else if (actionType == ACTION_EXTRACT)
                startElementExtract(uri, localName, qName, atts);
        }
        catch (Exception ex)
        {
            throw new SAXException(ex.getMessage());
        }
    }

    @Override
    public void endElement(String uri, String localName, String qName) throws SAXException
    {
        try
        {
            if (actionType == ACTION_INSERT)
                endElementInsert(uri, localName, qName);
            else if (actionType == ACTION_FORMAT)
                endElementFormat(uri, localName, qName);
            else if (actionType == ACTION_EXTRACT)
                endElementExtract(uri, localName, qName);
        }
        catch (Exception ex)
        {
            throw new SAXException(ex.getMessage());
        }
    }


    @Override
    public void characters(char[] ch, int start, int length) throws SAXException
    {
        try
        {
            if (actionType == ACTION_INSERT)
                charactersInsert(ch, start, length);
            else if (actionType == ACTION_FORMAT)
                charactersFormat(ch, start, length);
            else if (actionType == ACTION_EXTRACT)
                charactersExtract(ch, start, length);
        }
        catch (Exception ex)
        {
            throw new SAXException(ex.getMessage());
        }
    }

    public void startElementInsert(String uri, String localName, String qName, Attributes atts) throws Exception
    {
        if (qName.equals("TEI"))
            endCommentWithLineFeed = false;      // comments before root element need line feed, those after root element do not.

        bw.write("<");
        bw.write(qName);
        int numatts = atts.getLength();
        for (int i=0; i<numatts; i++)
        {
            String name = atts.getQName(i);
            String value = atts.getValue(i);
            bw.write(String.format(" %s=\"%s\"", name, value));
        }
        bw.write(">");
    }

    public void endElementInsert(String uri, String localName, String qName) throws Exception
    {
        bw.write(String.format("</%s>", qName));
    }

    public void charactersInsert(char[] ch, int start, int length) throws Exception
    {
        String s = new String(ch, start, length);
        s = s.replace("&", "&amp;");
        bw.write(s);
    }

    public void startElementFormat(String uri, String localName, String qName, Attributes atts) throws Exception
    {
        if (qName.equals("origin"))
            qName = "origin";
        pi.pushElement(qName, atts);
    }

    public void endElementFormat(String uri, String localName, String qName) throws Exception
    {
        pi.popElement(qName);
    }

    public void charactersFormat(char[] ch, int start, int length) throws Exception
    {
        pi.append(new String(ch, start, length));
    }

    public void startElementExtract(String uri, String localName, String qName, Attributes atts) throws Exception
    {
        pushPath(qName);
        String thepath = getPath();

        Integer cnt = allpaths.get(thepath);
        if (cnt == null)
            cnt = new Integer(0);
        allpaths.put(thepath, new Integer(cnt.intValue() + 1));

        if (thepath.startsWith(extractionPath))
        {
            bw.write("<");
            bw.write(qName);
            int numatts = atts.getLength();
            for (int i=0; i<numatts; i++)
            {
                String name = atts.getQName(i);
                String value = atts.getValue(i);
                bw.write(String.format(" %s=\"%s\"", name, value));
            }
            bw.write(">");
        }
    }

    public void endElementExtract(String uri, String localName, String qName) throws Exception
    {
        if (getPath().startsWith(extractionPath))
        {
            bw.write(String.format("</%s>", qName));
        }
        if (getPath().equals(extractionPath))
            bw.write("\n");
        popPath();
    }

    public void charactersExtract(char[] ch, int start, int length) throws Exception
    {
        if (getPath().startsWith(extractionPath))
            bw.write(new String(ch, start, length));
    }

    @Override
    public void comment(char[] ch, int start, int length) throws SAXException
    {
        try
        {
            boolean displayComment = false;
            if (actionType == ACTION_INSERT)
                displayComment = true;
            else if (actionType == ACTION_EXTRACT && getPath().equals(extractionPath))
                displayComment = true;
            if (displayComment)
            {
                bw.write("<!--");
                bw.write(new String(ch, start, length));
                bw.write("-->");
                if (endCommentWithLineFeed)
                    bw.write("\n");
            }
        }
        catch (Exception ex)
        {
            throw new SAXException(ex.getMessage());
        }
    }

    @Override
    public void processingInstruction(String target, String data) throws SAXException
    {
        if (bw == null || actionType == ACTION_FORMAT)
            return;

        if (extractionPath != null && !getPath().equals(extractionPath))
            return;

        try
        {
            bw.write("<?");
            bw.write(target);
            bw.write(" ");
            bw.write(data);
            bw.write("?>\n");
        }
        catch (Exception ex)
        {
            throw new SAXException(ex.getMessage());
        }
    }

    @Override
    public void ignorableWhitespace(char[] ch, int start, int length) throws SAXException
    {
    }



    private void reserveEntries()
    {
        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msIdentifier.collection");
        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msIdentifier.altIdentifier[type='*'].collection[type='desc']");

        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msIdentifier.idno");

        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msIdentifier.repository");

        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msIdentifier.institution");

        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msIdentifier.settlement");
        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msIdentifier.region");
        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msIdentifier.country");

        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msIdentifier.altIdentifier[type='catalog']");
        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msIdentifier.altIdentifier[type='catalog'].idno");

        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msContents.msItemStruct:defective");

        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msContents.msItemStruct.locus:from");
        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msContents.msItemStruct.locus:to");
        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msContents.msItemStruct.locus");

        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msContents.msItemStruct.author");
        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msContents.msItemStruct.author.persName");

        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msContents.msItemStruct.title[type='main']");

        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msContents.msItemStruct.title[type='part']");

        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msContents.msItemStruct.rubric[type='cover']");
        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msContents.msItemStruct.rubric[type='cover'].locus:from");
        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msContents.msItemStruct.rubric[type='*']");

        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msContents.msItemStruct.incipit");
        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msContents.msItemStruct.incipit.locus:from");
        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msContents.msItemStruct.incipit.locus:to");
        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msContents.msItemStruct.incipit.locus");
        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msContents.msItemStruct.incipit.bibl.biblScope");

        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msContents.msItemStruct.explicit");
        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msContents.msItemStruct.explicit.locus:from");
        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msContents.msItemStruct.explicit.locus:to");
        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msContents.msItemStruct.explicit.locus");
        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msContents.msItemStruct.explicit.bibl.biblScope");

        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msContents.msItemStruct.finalRubric");
        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msContents.msItemStruct.finalRubric.locus");
        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msContents.msItemStruct.finalRubric.locus:from");

        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msContents.msItemStruct.colophon");
        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msContents.msItemStruct.colophon.locus:from");

        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msContents.msItemStruct.filiation");

        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msContents.msItemStruct.witDetail.secFol");
        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msContents.msItemStruct.witDetail.secFol.locus");
        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msContents.msItemStruct.witDetail.secFol.bibl");

        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msContents.msItemStruct.witDetail.signatures");

        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msContents.msItemStruct.textLang");

        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.objectDesc:form");

        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.objectDesc.supportDesc.support.material");

        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.objectDesc.supportDesc.support.watermark");

        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.objectDesc.supportDesc.extent");
        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.objectDesc.supportDesc.extent.measure");
        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.objectDesc.supportDesc.extent.measure:unit");
        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.objectDesc.supportDesc.extent.measure:quantity");

        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.objectDesc.supportDesc.extent.dimensions:scope");
        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.objectDesc.supportDesc.extent.dimensions:type");
        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.objectDesc.supportDesc.extent.dimensions:unit");

        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.objectDesc.supportDesc.extent.dimensions.height");
        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.objectDesc.supportDesc.extent.dimensions.height");

        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.objectDesc.supportDesc.foliation");
        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.objectDesc.supportDesc.foliation.formula");

        pi.put("TEI.facsimile.graphic:url");

        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.objectDesc.supportDesc.collation");
        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.objectDesc.supportDesc.collation.formula");

        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.objectDesc.supportDesc.condition");

        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.accMat");

        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.objectDesc.layoutDesc");
        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.objectDesc.layoutDesc.layout");
        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.objectDesc.layoutDesc.layout:columns");
        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.objectDesc.layoutDesc.layout:ruledLines");
        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.objectDesc.layoutDesc.layout:writtenLines");

        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.handDesc.summary");

        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.handDesc.handNote");
        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.handDesc.handNote:scribe");


        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.additions");

        // for history
        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.history.origin");
        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.history.provenance");
        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.history.acquisition");

        // for subject headings
        pi.put("TEI.teiHeader.profileDesc.textClass.keywords[scheme='#LCSH'].list.item");

        pi.put("TEI.teiHeader.profileDesc.textClass.catRef:target");
        pi.put("TEI.teiHeader.profileDesc.textClass.catRef:scheme");

        pi.put("TEI.text");

        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.additional.adminInfo.recordHist.change");
        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.additional.adminInfo.recordHist.change:when");
        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.additional.adminInfo.recordHist.change:who");
        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.additional.adminInfo.recordHist.change.date");
        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.additional.adminInfo.recordHist.change.persName");
    }


    public void writeRow(StringBuilder sb, String value)
    {
        sb.append(value);
    }

    public void writeRow(StringBuilder sb, String type, String value)
    {
        sb.append(String.format(rowTemplate, type, value));
    }


    class Handler_msIdentifierStruct implements ElementHandler
    {

        public void process(String path)
        {
           manuscriptIdentifiers.append("<p><b>Manuscript Identifiers</b></p>\n");

           manuscriptIdentifiers.append("<table>\n");

            // Collection
            String val = pi.getTrimmedValue("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msIdentifier.collection");
            writeRow(manuscriptIdentifiers, "Collection", val);
            //val = pi.getTrimmedValue("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msIdentifier.altIdentifier[type='*'].collection[type='desc']");

            // Item
            val = pi.getTrimmedValue("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msIdentifier.idno");
            writeRow(manuscriptIdentifiers, "Item", val);

            // Repository
            val = pi.getTrimmedValue("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msIdentifier.repository");
            writeRow(manuscriptIdentifiers, "Repository", val);

            // Institution
            val = pi.getTrimmedValue("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msIdentifier.institution");
            writeRow(manuscriptIdentifiers, "Institution", val);

            // Location
            val = String.format("%s, %s, %s",
                        pi.getTrimmedValue("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msIdentifier.settlement"),
                        pi.getTrimmedValue("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msIdentifier.region"),
                        pi.getTrimmedValue("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msIdentifier.country"));
            writeRow(manuscriptIdentifiers, "Location", val);

            // Catalog
            val = pi.getTrimmedValue("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msIdentifier.altIdentifier[type='catalog']");
            writeRow(manuscriptIdentifiers, "Catalog", val);

            // Item
            val = pi.getTrimmedValue("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msIdentifier.altIdentifier[type='catalog'].idno");
            writeRow(manuscriptIdentifiers, "Item", val);

           manuscriptIdentifiers.append("</table>\n");
        }
    }


    class Handler_msItemStruct implements ElementHandler
    {
        public void process(String path)
        {
                String val = pi.getTrimmedValue("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msContents.msItemStruct:defective");
                boolean complete = val.toLowerCase().equals("false");
                works.append(String.format("<p><b>Work %d</b> (%s)</p>",
                            workNumber++, complete ? "complete" : "incomplete"));

                works.append("<table>\n");

                // Locus
                writeRow(works, "Locus", pi.getTrimmedValue("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msContents.msItemStruct.locus:from"));

                // Author
                val = pi.getTrimmedValue("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msContents.msItemStruct.author");
                if (val.length() == 0)
                    val = pi.getTrimmedValue("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msContents.msItemStruct.author.persName");

                writeRow(works, "Author", val);

                // Title
                val = pi.getTrimmedValue("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msContents.msItemStruct.title[type='main']");
                writeRow(works, "Title", val);

                // Part
                val = pi.getTrimmedValue("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msContents.msItemStruct.title[type='part']");
                writeRow(works, "Part", val);

                // Cover Rubric
                val = pi.getTrimmedValue("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msContents.msItemStruct.rubric[type='cover']");
                writeRow(works, "Cover rubric", val);

                // Incipit
                val = pi.getTrimmedValue("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msContents.msItemStruct.incipit.locus:from");
                if (val.length() > 0)
                    val = String.format("f.<%s> ", val);

                val += pi.getTrimmedValue("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msContents.msItemStruct.incipit");
                writeRow(works, "Incipit", val);

                // Explicit
                val = pi.getTrimmedValue("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msContents.msItemStruct.explicit.locus:from");
                if (val.length() > 0)
                    val = String.format("f.<%s> ", val);

                val += pi.getTrimmedValue("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msContents.msItemStruct.explicit");
                writeRow(works, "Explicit", val);

                // Final rubric
                val = pi.getTrimmedValue("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msContents.msItemStruct.finalRubric.locus:from");
                if (val.length() > 0)
                    val = String.format("f.<%s> ", val);

                val += pi.getTrimmedValue("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msContents.msItemStruct.finalRubric");
                writeRow(works, "Final rubric", val);

                // Colophon
                val = pi.getTrimmedValue("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msContents.msItemStruct.colophon.locus:from");
                if (val.length() > 0)
                    val = String.format("f.<%s> ", val);

                val += pi.getTrimmedValue("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msContents.msItemStruct.colophon");
                writeRow(works, "Colophon", val);

                // Filiation
                val = pi.getTrimmedValue("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msContents.msItemStruct.filiation");
                writeRow(works, "Filiation", val);

                // Second folio
                writeRow(works, "Second folio", "Don't understand");

                // Signatures
                writeRow(works, "Signatures", "Don't understand");

                // Language
                val = pi.getTrimmedValue("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msContents.msItemStruct.textLang");
                writeRow(works, "Language", val);

                works.append("</table>\n");
        }
    }

    class Handler_physDesc implements ElementHandler
    {
        public void process(String path)
        {
           physicalDescription.append("<p><b>Physical description</b></p>\n");

           physicalDescription.append("<table>\n");

            // Form
            String val = pi.getTrimmedValue("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.objectDesc:form");
            writeRow(physicalDescription, "Form", val);

            // Material
            val = pi.getTrimmedValue("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.objectDesc.supportDesc.support.material");
            writeRow(physicalDescription, "Material", val);

            // Watermark
            val = pi.getTrimmedValue("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.objectDesc.supportDesc.support.watermark");
            writeRow(physicalDescription, "Watermark", val);

            // Extent
            val = pi.getTrimmedValue("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.objectDesc.supportDesc.extent.measure:quantity");
            writeRow(physicalDescription, "Extent", val);

            // Foliation
            val = String.format("%s<br/>%s",
                    pi.getTrimmedValue("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.objectDesc.supportDesc.foliation"),
                    pi.getTrimmedValue("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.objectDesc.supportDesc.foliation.formula"));
            writeRow(physicalDescription, "Foliation", val);

            // Collation
            val = String.format("%s<br/>%s",
                    pi.getTrimmedValue("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.objectDesc.supportDesc.collation"),
                    pi.getTrimmedValue("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.objectDesc.supportDesc.collation.formula"));
            writeRow(physicalDescription, "Collation", val);

            // Condition
            val = pi.getTrimmedValue("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.objectDesc.supportDesc.condition");
            writeRow(physicalDescription, "Condition", val);

            // Binding
            val = pi.getTrimmedValue("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.bindingDesc");
            pd.parse(val);
            writeRow(physicalDescription, pd.getHTML());

            // Seal
            val = pi.getTrimmedValue("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.sealDesc");
            pd.parse(val);
            writeRow(physicalDescription, pd.getHTML());

            // Accompanying matter
            val = pi.getTrimmedValue("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.accMat");
            writeRow(physicalDescription, "Accompanying matter", val);

            physicalDescription.append("</table>\n");
        }
    }


    class Handler_handNote implements ElementHandler
    {

        public void process(String path)
        {
            String val = pi.getTrimmedValue("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.handDesc.handNote");
            pd.parse(val);
            handNotes.add(String.format("<p>%s</p>", pd.getHTML()));
        }

    }



    class Handler_handDesc implements ElementHandler
    {

        public void process(String path)
        {
            hands.append("<p><b>Hands</b></p>\n");

            hands.append("<table>\n");

            String val = pi.getTrimmedValue("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.handDesc.summary");
            writeRow(hands, "Summary", val);

            int n = 1;
            for (String handNote : handNotes)
            {
                writeRow(hands, String.format("Hand %d", n++), handNote);
            }

            val = pi.getTrimmedValue("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.additions");
            pd.parse(val);
            writeRow(hands, "Additions", pd.getHTML());

            hands.append("</table>");
            handNotes.clear();
        }

    }


    class Handler_decoDesc implements ElementHandler
    {
        public void process(String path)
        {
            decoration.append("<p><b>Decoration</b></p>\n");

            decoration.append("<table>\n");

            String val = pi.getTrimmedValue("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.decoDesc.decoNote[type='color']");
            pd.parse("<root>"+val+"</root>");
            writeRow(decoration, "Color", pd.getHTML());

            val = pi.getTrimmedValue("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.decoDesc.decoNote[type='border']");
            pd.parse("<root>"+val+"</root>");
            writeRow(decoration, "Border", pd.getHTML());

            val = pi.getTrimmedValue("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.decoDesc.decoNote[type='illustration']");
            pd.parse("<root>"+val+"</root>");
            writeRow(decoration, "Illustration", pd.getHTML());

            hands.append("</table>");
        }
    }

    class Handler_history implements ElementHandler
    {
        public void process(String path)
        {
            history.append("<p><b>History</b></p>\n");

            history.append("<table>\n");

            String val = pi.getTrimmedValue("TEI.teiHeader.fileDesc.sourceDesc.msDesc.history.origin");
            pd.parse(val);
            writeRow(history, "Origin", pd.getHTML());

            val = pi.getTrimmedValue("TEI.teiHeader.fileDesc.sourceDesc.msDesc.history.provenance");
            pd.parse(val);
            writeRow(history, "Provenance", pd.getHTML());

            val = pi.getTrimmedValue("TEI.teiHeader.fileDesc.sourceDesc.msDesc.history.acquisition");
            pd.parse(val);
            writeRow(history, "Acquisition", pd.getHTML());

            history.append("</table>");
        }
    }

    class Handler_profileDescItem implements ElementHandler
    {
        public void process(String path)
        {
            String val = pi.getTrimmedValue("TEI.teiHeader.profileDesc.textClass.keywords[scheme='#LCSH'].list.item");
            subjectItems.add(val);
        }
    }

    
    class Handler_profileDesc implements ElementHandler
    {
        public void process(String path)
        {
            subjectHeadings.append("<p><b>Subject headings</b></p>\n");

            subjectHeadings.append("<table>\n");

            for (String item : subjectItems)
            {
                writeRow(subjectHeadings, "SubjectLC", item);
            }

            String val = String.format("%s (%s)",
                pi.getTrimmedValue("TEI.teiHeader.profileDesc.textClass.catRef:target"),
                pi.getTrimmedValue("TEI.teiHeader.profileDesc.textClass.catRef:scheme"));

            writeRow(subjectHeadings, "SubjectSL", val);

            subjectHeadings.append("</table>");
        }
    }

    class Handler_facsimileGraphic implements ElementHandler
    {
        public void process(String path)
        {
            String val = pi.getTrimmedValue("TEI.facsimile.graphic:url");
            facsimileItems.add(val);
        }
    }

    class Handler_facsimile implements ElementHandler
    {
        public void process(String path)
        {
            subjectHeadings.append("<p><b>Facsimile</b></p>\n");

            StringBuilder sb = new StringBuilder();
            for (String item : facsimileItems)
            {
                sb.append(String.format("<a href='%s'>%s</a><br/>\n", item, item));
            }

            subjectHeadings.append(sb.toString());
        }
    }

    class Handler_transcription implements ElementHandler
    {
        public void process(String path)
        {
            transcription.append("<b>Transcription</b>\n");

            String val = pi.getTrimmedValue("TEI.text");
            pd.parse(val);            
            transcription.append(pd.getHTML());
        }
    }


    class Handler_change implements ElementHandler
    {
        public void process(String path)
        {
            ChangeInfo ci = new ChangeInfo(
                        pi.getTrimmedValue("TEI.teiHeader.fileDesc.sourceDesc.msDesc.additional.adminInfo.recordHist.change.persName"),
                        pi.getTrimmedValue("TEI.teiHeader.fileDesc.sourceDesc.msDesc.additional.adminInfo.recordHist.change.date"),
                        pi.getTrimmedValue("TEI.teiHeader.fileDesc.sourceDesc.msDesc.additional.adminInfo.recordHist.change")
                        );

           changeItems.add(ci);
        }
    }


    class Handler_changes implements ElementHandler
    {
        public void process(String path)
        {
            changes.append("<p><b>Changes</b></p>\n");
            changes.append("<table>\n");
            changes.append("<tr><th>Person</th><th>Date</th><th>Change</th></tr>\n");
            String template = "<tr><td>%s</td><td>%s</td><td>%s</td></tr>\n";
            for (ChangeInfo ci : changeItems)
            {
                changes.append(String.format(template, ci.person, ci.date, ci.comment));
            }

            changes.append("</table>\n");
        }

    }

    class Handler_default implements ElementHandler
    {
        public void process(String path)
        {
            if (path != null)
                return;

            String buffer = pi.getTrimmedValue(path);
            if (buffer != null)
                System.out.println(String.format("%s:---->%s", path, pi.getTrimmedValue(path)));
            else
                System.out.println(String.format("%s:---->null", path));

        }
    }


    class Rewrite_default implements TagHandler
    {
        boolean breakOnSingleDanda = false;

        // Methods for TagHandler interface
        public String getStartTag(ArrayList<String> context, Attributes atts)
        {
            int top = context.size() - 1;
            String tag = context.get(top);
            if (tag.equals("p"))
                return "<p>";
            else if (tag.equals("note"))
            {
                String resp = atts.getValue("resp");
                if (resp != null)
                {
                    return String.format("<br/>Note: (%s) ", resp);
                }
                else
                    return "<br/>Note: ";
            }
            else if (tag.equals("bindingDesc"))
                return "<tr><td>Binding:</td><td> ";
            else if (tag.equals("condition"))
                return "<tr><td>Condition</td><td>";
            else if (tag.equals("add"))
                return "(";
            else if (tag.equals("corr"))
                return "&lt;";
            else if (tag.equals("del"))
                return "[";
            else if (tag.equals("sic"))
                return "[";
            else if (tag.equals("unclear"))
                return "&dagger;";
            else if (tag.equals("quote"))
                return "&ldquo;";
            else if (tag.equals("desc"))
                return "(";
            else if (tag.equals("l"))
                return "<br/>";
            else if (tag.equals("ab"))
            {
                breakOnSingleDanda = true;
                return "<br/>";
            }
            else if (tag.equals("pb"))
            {
                String type = atts.getValue("type");
                if (type != null && type.equals("folio"))
                {
                    String n = atts.getValue("n");
                    if (n == null)
                        n = "?";
                    return String.format("<br/><br/>f. %s:<br/>", n);
                }
            }
            else if (tag.equals("div"))
            {
                String type = atts.getValue("type");
                if (type != null)
                {
                    if (type.equals("mUla"))
                        return String.format("%s:<br/>", "m\u016b");
                    else if (type.equals("commentary"))
                        return "Commentary:<br/>";
                    else if (type.equals("subcommentary"))
                        return "Subcommentary:<br/>";
                    else
                    {
                        String n = atts.getValue("n");
                        return String.format("%s %s:<br/>", type, n);
                    }
                }
            }
            else if (tag.equals("bibl"))
            {
                String type = atts.getValue("type");
                if (type != null && !type.equals("verse"))
                    ignoreUntil = context.size();
                else
                    return "(";
            }

            return "";
        }

        public String getFormat(ArrayList<String> context, Attributes atts)
        {
            return null;
        }

        public boolean isPrintable()
        {
            return ignoreUntil == -1;
        }

        public String getEndTag(ArrayList<String> context)
        {
            if (ignoreUntil == context.size())
            {
                ignoreUntil = -1;
                return "";
            }
            
            int top = context.size() - 1;
            String tag = context.get(top);
            if (tag.equals("p"))
                return "</p>";
            else if (tag.equals("note"))
                return "<br/>";
            else if (tag.equals("bindingDesc"))
                return "</td></tr>\n";
            else if (tag.equals("condition"))
                return "</td></tr>\n";
            else if (tag.equals("add"))
                return ")";
//            else if (tag.equals("lg"))
//                return "<br/>";
            else if (tag.equals("corr"))
                return "&gt;";
            else if (tag.equals("del"))
                return "]";
            else if (tag.equals("sic"))
                return "]";
            else if (tag.equals("unclear"))
                return "&dagger;";
            else if (tag.equals("quote"))
                return "&rdquo;";
            else if (tag.equals("desc"))
                return ")";
            else if (tag.equals("ab"))
                breakOnSingleDanda = false;
            else if (tag.equals("bibl"))
                return ") ";

            return "";
        }

        public String fixup(String s)
        {
            if (breakOnSingleDanda)
                s = s.replace("| ", "| <br/>");

            return s;
        }
    }
}
